/*! \file  thread.c
\brief Implantation de la gestion des threads
*/

#include "thread.h"
#include "mask.h"
#include "interrupt.h"
#include "countdown.h"
#include "contextswitch.h"
#include "Cstack.h"
#include "mem.h"
#include	<errno.h>

/*! \ingroup IGthread
\brief Descripteur de  threads 

	Le champ FredThreadDesc::err est la sauvegarde de la variable globale errno qui contient le code d'erreur du programme. On la sauve et la restaure  chaque commutation de thread.

*/
typedef struct  FredThreadDesc
{
  FredCELL;   				/*!< pour mettre les thread en file*/
  char * stack;  			/*!<  adr zone pile */
  int      stacksz;		/*!<  taille  de pile  */
  FredRunnable entry; /*!< procdure appele au lancement du thread*/
  void *      val;    /*!<  argument  */
  void **			resref; /*!< adresse resultat */
	int err;					/*!< sauvegarde code erreur errno*/
  FredContext   ctx  ; /*!< contexte registres processeur */         
} FredThreadDesc;


 
/*! \ingroup IGthread
\brief Compteur du nombre de threads crs non termins

	Seuls les threads crs par FredThreadSpawn() sont compts. Ce compteur sert  dtecter la terminaison du programme FRED (cf FredWaitTerminate() ).

*/
	int volatile nbFredThreads; /* nombre de threads vivants */

/*! \ingroup IGthread
\brief Le thread actif

*/
 	FredThread   current ;
 
/*! \ingroup IGthread
\brief File d'attente des threads prts

	Elle n'est jamais vide du fait de l'existence du chien de garde #watchdog
*/
 	FredQueue  readyQ; 


/*! \ingroup IGthread
 \brief File d'attente des threads morts dont la pile est  librer (rserve  #leon le nettoyeur)
	Un thread terminant s'inscrit dans cette file et rveille #leon s'il n'est pas dja rveill ( cf leonCode() ).

*/
	FredQueue  deadQ ;
/*! \ingroup IGthread
 \brief  Le nettoyeur leon 

	Comme un thread terminant ne peut librer la pile dans laquelle il s'excute, ce travail est assur par un thread spcifique de nettoyage : #leon . Il attend qu'un thread mourant le rveille dans une file d'attente spcifique #leonQ . Il trouve les threads dcds dans la file #deadQ ( cf leonCode() ) Il assure aussi la dtection de terminaison.

*/
	FredThread  leon; 

/*! \ingroup IGthread
 \brief File d'attente  de #leon
*/
	FredQueue  leonQ ; /* file d'attente de leon*/

/*! \ingroup IGthread
 \brief File d'attente de terminaison (rserve au #mainThread )
*/
	FredQueue endQ;   
/*! \brief Le thread associ  la procdure main() 
*/
	FredThread   mainThread;   
/*! \ingroup IGthread
\brief Descripeur du thread associ  la procdure main() 
*/
	FredThreadDesc mainThreadBlock; 

/*! \ingroup IGthread
\brief Thread chien de garde

	Son rle est d'viter une file #readyQ  vide. Il ne fait que passer la main. ( cf. watchDogCode() ).
*/
	FredThread  watchdog; 

	






/*! \ingroup IGthread
 \brief  Le code du nettoyeur leon 

	Il est en charge de la libration de mmoire des threads morts et de la dtection de terminaison.

\htmlonly <hr WIDTH="100%"> \endhtmlonly 
\htmlonly <hr WIDTH="100%"> \endhtmlonly 
\b Implantation

leon attend qu'un thread mourant le rveille dans sa file d'attente spcifique #leonQ . Il trouve les threads dcds dans la file #deadQ et libre la mmoire.
Il dcrmente le compte de threads #nbFredThreads et dtecte la terminaison. Dans ce cas, il rveille le thread principal si ncessaire.

*/

void *  leonCode(void * x)
{
	FredThread t;	
	while(1){
		FredMaskOn();/* dbut d'exclusion mutuelle */
		/* librer la mmoire des threads dfunts */
		while((t=FredQueueGetFirst(&deadQ))){
			free(t->stack);free(t);
			nbFredThreads--;
		}
		/* si terminaison, on rveille le thread l'attendant */
		if(nbFredThreads==2)FredThreadWakeUp(&endQ);
		/* leon attend de nouveaux dcs */
		FredThreadSuspend(&leonQ);
		FredMaskOff();
	}
	/* on ne revient jamais la */
  FredAbort(" erreur fatale dans leon");
}

/*! \ingroup IGthread
 \brief  Initialisation de leon le nettoyeur

\htmlonly <hr WIDTH="100%"> \endhtmlonly 
\htmlonly <hr WIDTH="100%"> \endhtmlonly 
\b Implantation
	On lance leon aprs avoir initialiser les 2 files #leonQ et #deadQ .

*/


void LeonInit()
{
	FredQueueInit(&deadQ); /*  file vide */
	FredQueueInit(&leonQ); /*  file vide */
	leon=FredThreadSpawn(20000,leonCode,(void*)0,(void**)0);
}
/****************************************************************
*                                                              
* Chien de garde :
* 
*			- S'il n'y a plus de thread, termine le programme
*			- s'il n'y a plus de thread prts, il tourne.
*				Ce peut tre un  deadlock s'il n'y a pas de thread en attente 
*					de signaux externes                                                    
*****************************************************************
*****************************************************************/

/*! \ingroup IGthread
 \brief  Le code du chien de garde watchdog

Il vite d'avoir une file #readyQ vide  traiter.
	
\htmlonly <hr WIDTH="100%"> \endhtmlonly 
\htmlonly <hr WIDTH="100%"> \endhtmlonly 
\b Implantation
	C'est un boucle infinie qui ne fait que passer la main  un autre thread prt.
*/


void * watchDogCode(void * x)
{
	while(1){
		FredThreadYield();
		/* attraper les interruptions externes */
	}
	/* on ne revient jamais la */
  FredAbort(" erreur fatale dans chien de garde");
}

/*! \ingroup IGthread
 \brief  Initialisation du chien de garde watchdog

\htmlonly <hr WIDTH="100%"> \endhtmlonly 
\htmlonly <hr WIDTH="100%"> \endhtmlonly 
\b Implantation
*/

void WatchDogInit()
{
	watchdog=FredThreadSpawn(20000,watchDogCode,(void*)0,(void**)0);
}

/*! \ingroup IGthread
 \brief  Valeur du quantum en microseconde 
*/
long int quantum; /* nombre de microsecondes */ 


/*! \ingroup IGthread
 \brief  Traitement d'interruption de fin de quantum

	\htmlonly <hr WIDTH="100%"> \endhtmlonly 
\htmlonly <hr WIDTH="100%"> \endhtmlonly 
\b Implantation
	On recycle en fin de #readyQ le thread courant. On choisit le suivant qu'on relance avec un nouveau dcomptage de quantum.

*/
void QuantumHandler(int i)
{
	FredThread old;
  /* interruption de fin de quantum 
		on est masqu contre l'it quantum*/
  /* printf("+"); */
	FredMaskOn(); /* dbut d'exclusion mutuelle */
	current->err=errno; /* sauver errno */
	old=current;
	FredQueuePutLast(&readyQ,current);
	current=FredQueueGetFirst(&readyQ);
	if(current!=old)FredContextSwitch(old->ctx,current->ctx);
	/* 
		on reviendra ici lors de son rveil :
		on relance le dcomptage du quantum
 	*/
	FredCountDownStart(quantum);
	errno=current->err;/* restaurer errno */
	FredMaskOff();/* fin d'exclusion mutuelle */
}

/*! \ingroup IGthread
 \brief  Initialisation du temps partag 

\htmlonly <hr WIDTH="100%"> \endhtmlonly 
\htmlonly <hr WIDTH="100%"> \endhtmlonly 
\b Implantation
	On initialise le quantum avec une valeur donne et on attache le traitement d'interruption de fin de quantum au dcompteur.

*/

void TimeSharingInit( long int uS)
{
	/* attach  une interuption un traitant d'IT */
	FredCountDownAttach(QuantumHandler);
	/* valeur du quantum */
	quantum=uS;
}


/***************************************************************
****************************************************************
****************************************************************
*                                                              
*   API threads
*
****************************************************************
*****************************************************************
*****************************************************************/


/***************************************************************/  
/*! \ingroup Gcond
\brief Suspendre le thread courant dans une file
\param q file o mettre le thread courant 

	Le thread courant est mis en queue de la file. Un nouveau thread est lu pour s'excuter.

\warning Cette procdure n'est utilisable qu'au sein d'une section en exclusion mutuelle. 


\htmlonly <hr WIDTH="100%"> \endhtmlonly 
\htmlonly <hr WIDTH="100%"> \endhtmlonly 
\b Implantation

	On doit arrter puis relancer un dcomptage.
*/
void FredThreadSuspend( FredQueue* q)
{
	FredThread old;	
	current->err=errno; /* sauver errno */
	FredCountDownStop(); /* Arrt dcompeur */
	/* le courant se  en queue de la file q */
	FredQueuePutLast(q,current); 
	/* 
		on choisit un nouveau courant : on doit garantir
		que la file readyQ n'est jamais vide : C'est le rle
		du watchdog
	*/
 	old=current; current=FredQueueGetFirst(&readyQ);
	FredContextSwitch(old->ctx,current->ctx);
	/* 
		on reviendra ici lors de son rveil :
		on relance le dcomptage du quantum
	*/
		FredCountDownStart(quantum);
		errno=current->err;/* restaurer errno */
}



/*! \ingroup Gcond
\brief Rveille un  thread bloqu
\param q queue dont- le premier thread est remis prt

	Le thread rveill  est mis en queue de la file des prts. C'est une opration nulle si la file est vide.

\warning Cette procdure n'est utilisable qu'au sein d'une section en exclusion mutuelle.

\htmlonly <hr WIDTH="100%"> \endhtmlonly 
\htmlonly <hr WIDTH="100%"> \endhtmlonly 
\b Implantation
*/


void FredThreadWakeUp( FredQueue* q)
{
		if(!FredQueueEmpty(q))FredQueuePutLast(&readyQ,FredQueueGetFirst(q));
}





/*! \ingroup IGthread
\brief Code initial des threads

	Comme on ne peut retourner de faon normale de la procdure initiale d'un thread et comme le retour de celle-ci est le retour standard gnr par le compilateur C, on utilise une procdure  qui encadre (harnache) le dmarrage et la fin d'un thread. 
	
\htmlonly <hr WIDTH="100%"> \endhtmlonly 
\htmlonly <hr WIDTH="100%"> \endhtmlonly 
\b Implantation

	
Ce harnais appelle la procdure initiale du thread, rcupre son rsultat et dlgue  #leon la libration de la pile sans oublier les masquages et dcomptages ncessaires.
*/


void FredThreadHarness()
{/* on dmarre masqu */
	FredThread old;
	/* lancer le dcomptage */
	FredCountDownStart(quantum);
	/* dmasquage */ 
	FredMaskOff(); 
	/* appel de la procdure initiale avec ou sans rcupration dec rsultat */
	if(current->resref){ /* on attend un rsultat */
			*(current->resref)=current->entry(current->val);

	}else{		
			current->entry(current->val);
	}
	/* arrt dcomptage */
	FredCountDownStop(); 
	/* dbut exclusion mutuelle */
	FredMaskOn();
	/* rveil de leon */
	FredThreadWakeUp(&leonQ);
	/* 
		un thread ne peut librer la mmoire
		dans laquelle il s'excute. Il dlgue cela 
		a leon  en s'inscrivant dans la
		file des threads dfunts
	*/
	FredThreadSuspend(&deadQ);
	/* on ne revient jamais la */
  FredAbort(" erreur fatale dans threadHarness");
}


/***************************************************************/  
/*! \ingroup Gthread
\brief Cration et lancement d'un thread.
\param stacksz taille de pile exige pour le thread
\param function fonction  excuter par le thread
\param arg   adresse de l'argument de la fonction
\param resref Si l'adresse resref est non nulle, on y mettra l'adresse du rsultat  de la fonction lorsque le thread cr terminera.
\return l'adresse du descripteur du thread cr ou FredTHREAD_NIL (0)

\htmlonly <hr WIDTH="100%"> \endhtmlonly 
\htmlonly <hr WIDTH="100%"> \endhtmlonly 
\b Implantation

On alloue la mmoire pour un descripteur de thread et la pile (ventuellement arrondi ). Si cette tape est franchi avec succs, on initialise le descripteur du thread ( #FredThreadDesc ) et son contexte.  On l'insre enfin dans la file des prts #readyQ en incrmentant le nombre de threads vivants #nbFredThreads .

*/

FredThread   FredThreadSpawn(int stacksz,
								FredRunnable function,
                void *arg,void ** resref)
{
	FredThread nw;
	/* allocation d'un descripteur de thread si possible */
	nw= FredNEW(FredThreadDesc) ;
  if(!nw) return nw;
  /* calcul taille de pile */
  nw->stacksz=FredSTACK_SIZE(stacksz); 
	/* allocation de la pile  si possible */
  if(!(nw->stack=(char*)FredAtomicAlloc(nw->stacksz))) { 
		/* plus de mmoire */
    FredDELETE( nw); return FredTHREAD_NIL;
	}
	/* initialisation du descripteur */
	nw->err=0; /* init errno  OK */
	nw-> entry=function; /* fonction  excuter */
	nw->val=arg; /* son argument */
  nw->resref=resref; /* o mettre le rsultat */
	/* initialisation du contexte processeur du thread */ 
	FredContextInit(nw->ctx,
			FredMAKE_SP(nw->stack,nw->stacksz),
			FredThreadHarness);
	FredMaskOn();/* dbut d'exclusion mutuelle */
	FredQueuePutLast(&readyQ,nw);/* rveil du nouveau thread */
	nbFredThreads++; /* un thread de plus !! */
	FredMaskOff();/* fin d'exclusion mutuelle */
	return nw;
}





/***************************************************************/  
/*! \ingroup Gthread
\brief Terminaison immdiate du thread courant
\param res adresse du rsultat du calcul du thread

\htmlonly <hr WIDTH="100%"> \endhtmlonly 
\htmlonly <hr WIDTH="100%"> \endhtmlonly 
\b Implantation
	
On rend le rsultat  l'endrioit dfini au mau moment de la cration du thread par FreedThreadSpawn() . Le thread est mis dans la file des dcds pour que sa mmoire soit librer par #leon
*/

void   FredThreadExit(void * res)
{
  FredThread old;
	FredCountDownStop();/* arrt dcomptage */
	/* mettre le rsultat  l'endroit demand */
  if(current->resref)*(current->resref)=res; 
	FredMaskOn();/* dbut d'exclusion mutuelle */
	/* rveil de leon */
	FredThreadWakeUp(&leonQ);
	/* 
		un thread ne peut librer la mmoire
		dans laquelle il s'excute. Il dlgue cela 
		a leon  en s'inscrivant dans la
		file des threads dfunts
	*/
	FredThreadSuspend(&deadQ);
	/* on ne revient jamais ici */
	FredAbort(" erreur fatale dans threadDie");
}
/***************************************************************/  
/*! \ingroup Gthread
\brief Le thread courant passe la main  un autre thread

\htmlonly <hr WIDTH="100%"> \endhtmlonly 
\htmlonly <hr WIDTH="100%"> \endhtmlonly 
\b Implantation
*/

void FredThreadYield( )
{
	FredThread old;
	FredMaskOn(); /* dbut d'exclusion mutuelle */
	current->err=errno; /* sauver errno */
  FredCountDownStop();/* arrt dcomptage */
	old=current;
	/* le courant se met en queue des prts */
	/* on tire un nouveau courant */
	FredQueuePutLast(&readyQ,current);
	current=FredQueueGetFirst(&readyQ);
	/* traiter le cas d'un unique thread actif : rien faire */
	if(current!=old)FredContextSwitch(old->ctx,current->ctx);
	/* 
		on reviendra ici lors de son rveil :
		on relance le dcomptage du quantum
 	*/
	FredCountDownStart(quantum);/* lancer dcomptage */
	errno=current->err;/* restaurer errno */
	FredMaskOff();/* fin d'exclusion mutuelle */
}

/***************************************************************/  
/*! \ingroup Gthread
\brief Rend l'identit du thread courant
\return l'adresse du descripteur du thread actif                        

\htmlonly <hr WIDTH="100%"> \endhtmlonly 
\htmlonly <hr WIDTH="100%"> \endhtmlonly 
\b Implantation

*/
FredThread	FredThreadCurrent()
{
	return current;
}
 

/***************************************************************/  

/**************************************************************
* Initialisation mainThread
**************************************************************/

/*! \ingroup IGthread
\brief Initialisation du pseudo thread principal

\htmlonly <hr WIDTH="100%"> \endhtmlonly 
\htmlonly <hr WIDTH="100%"> \endhtmlonly 
\b Implantation
	La procdure principale main() initialise la file d'attente de terminaison, son bloc descripteur et devient le thread courant. A partir de la, elle peut utiliser l'interface FRED comme tout thread normal (synchronisation etc..).
Elle peut attendre la terminaison des autres threads via la file #endQ (cf. FredWaitTerminate() ).
*/
void MainThreadInit(){
	/* file d'attente de disparition des autres threads */
	FredQueueInit(&endQ); /*  file vide */
	/* allocation decripteur */
	mainThread=&mainThreadBlock;
	/* thread actif */
	current=mainThread;
}



/*! \ingroup Ginit
\brief Initialisation de FRED

\htmlonly <hr WIDTH="100%"> \endhtmlonly 
\htmlonly <hr WIDTH="100%"> \endhtmlonly 
\b Implantation
	La procdure principale main() initialise les variables d'tat des diffrents modules de base, se transforme en un thread principal et lance le nettoyeur leon et le veilleur watchdog. 
*/
 void FredInitialize()
{
	FredMasksInit(); /* initialiser les masques de signaux */
	FredSignalsInit();/* initialiser les traitants de signaux */
	nbFredThreads=0;/* aucun thread n'existe */
	FredQueueInit(&readyQ); /* file vide */
	TimeSharingInit(250); /* initialiser le temps partag */
	MainThreadInit(); /* forger un pseudo thread principal */
	LeonInit(); /* crer le nettoyeur */
	WatchDogInit(); /* crer le chien de garde */
}
/***************************************************************/  
/*! \ingroup Ginit
\brief  Attente de fin d'excution de tous les threads crs

\htmlonly <hr WIDTH="100%"> \endhtmlonly 
\htmlonly <hr WIDTH="100%"> \endhtmlonly 
\b Implantation
Le thread principal se bloque s'il y d'autres threads que le watchdog et leon. Sinon il termine.
*/
 void  FredWaitTerminate()
{ int nonfini=1;
  while(nonfini){
		FredMaskOn();/* dbut d'exclusion mutuelle */
		/* Il y a 2 threads de servitude : leon et watchdog */
		if(nbFredThreads>2) /* il y a d'autres threads */
			/* suspension dans la file d'attente de fin */
			FredThreadSuspend(&endQ);
		else	/* il n'y a plus que  leon et watchdog */
			nonfini=0;
		FredMaskOff();/* fin d'exclusion mutuelle */
	}
}
/***************************************************************/  

/*! \ingroup Ginit
\brief Arrt brutal du programme et impression d'un message d'erreur
\param mess message d'erreur fatale


\htmlonly <hr WIDTH="100%"> \endhtmlonly 
\htmlonly <hr WIDTH="100%"> \endhtmlonly 
\b Implantation
*/
/***************************************************************/  
void  FredAbort(char *mess )
{
  printf(" \n FATAL  ERROR : ");
  printf(mess);
  printf("\n");
  exit(1); /* fin du programme */
}


