/*! \file phil.h
\brief Inclusion de l'interface PHIL
*/
#ifndef _PHIL_H
#define _PHIL_H




/*! \ingroup Gpthread
\brief Le type d'une procdure excutable par un thread 
*/

typedef void * (*PhilRunnable)(void *);

/*! \ingroup Gpthread
\brief  L'identit d'un thread est un pointeur sur son descripteur
*/
typedef struct PhilThreadDesc* PhilThread;
/*! \ingroup Gpthread
\brief Cration d'un thread 
\param type le thread est joignable  0 ou dtach 1
\param priority >=0 dfinit  la priorit du thread cr. 0 est la priorit maximum. 
\param stacksize taille de pile demande
\param start_routine procdure initiale du thread
\param arg paramtre initial
\return l'identit du thread ou 0 si echec de cration

Un thread dtach ne peut tre joint. Il est ignor lors de la dtection de terminaison qui ne prend en compte que les threads joignables.

*/



extern PhilThread  PhilThreadCreate(int type, int priority, int stacksize,  PhilRunnable start_routine, void * arg); 

/*!\ingroup Gpthread
 \brief Joindre un thread 
\param th le thread joint 
\param return_val o mettre l'adresse du rsultat du thread joint. Si 0, le rsultat est ignor.
\return un code d'erreur :
	- 0 OK 
	- 1 le thread joint est inexistant 

Un thread dtach ne peut tre join. Il est ignor lors de la dtection de terminaison qui ne prend en compte que les threads joignables.

*/


extern int PhilThreadJoin(PhilThread th, void **return_val);

/*!\ingroup Gpthread
 \brief Terminer le thread courant
\param return_val  adresse de la valeur rsultat.

*/

extern  void PhilThreadExit(void *return_val);

/*!\ingroup Gpthread
 \brief  Obtenir l'identit du thread courant 
\return  identit du thread courant 

*/
extern PhilThread  PhilThreadCurrent();


/*!\ingroup Gpthread
 \brief  Passer la main  un autre thread 


	Le passage de main ne peut se faire qu'au sein de threads de mme priorit ou de priorits suprieures au thread courant.

*/
extern void  PhilThreadYield();

/*! \ingroup Gpmutex
\brief  Descripteur du verrou.

L'identit d'un verrou est un pointeur sur son descripteur
*/
typedef struct PhilMutex
{
/*  dfinir */
} PhilMutex;


/*!\ingroup Gpmutex
\brief  Initialiser un verrou
\param  verrou adresse du verrou
\return 0 OK  1 verrou inexistant (verrou==0)
*/

extern int  PhilMutexInit( PhilMutex* verrou);

/*!\ingroup Gpmutex
\brief  Destruction du verrou
\param  verrou adresse du verrou
\return un code d'erreur :
	- 0 OK  
	- 1 le verrou est inexistant ou non initialis	
	- 2 Le thread n'est pas le propritaire du verrou
	- 3 des threads attendent le verrou.

	On ne peut dtruire qu'un verrou dont on est propritaire. S'il y a des threads en attente, ils sont rveills et la destruction leur est signale.

*/
 extern int  PhilMutexDestroy(PhilMutex* verrou);


/*!\ingroup Gpmutex
\brief  Prendre le  verrou
\param  verrou adresse du verrou
\return un code d'erreur :
	- 0 OK 
	- 1 le verrou est inexistant ou non initialis
	- 2 le verrou a t dtruit ( PhilMutexDestroy() ) durant l'attente 

	Le verrou peut se verrouiller de faon rcursive.Il est allou au thread + prioritaire qui devient propritaire. 
*/


extern int  PhilMutexLock(PhilMutex* verrou);

/*!\ingroup Gpmutex
\brief  Essai de fermeture du verrou
\param  verrou adresse du verrou
\return un code d'erreur :
	- 0 OK  verrou ferm
	- 1 le verrou est inexistant ou non initialis
	- 2 echec de fermeture

	Le verrou peut se verrouiller de faon rcursive.Il est allou au thread + prioritaire qui devient propritaire. Il n'y a jamais blocage. 
*/

 extern int  PhilMutexTryLock(PhilMutex* verrou);

/*!\ingroup Gpmutex
\brief  Ouverture du verrou
\param  verrou adresse du verrou
\return un code d'erreur :
	- 0 OK  
	- 1 le verrou est inexistant ou non initialis
	- 2 Le thread n'est pas le propritaire du verrou


	Comme le verrou peut tre verrouill de faon rcursive, il ncessite le mme nombre de dverrouillages que de verrouillages pour tre ouvert.
Par ailleurs, on ne peut dverrouiller qu'un verrou dont on est propritaire. 
*/
 extern int  PhilMutexUnlock(PhilMutex* verrou);


/*! \ingroup Gpcond
\brief  Descripteur de condition.

L'identit d'une condition est un pointeur sur son descripteur
*/
typedef struct PhilCond
{
/*  dfinir */
} PhilCond;


/*!\ingroup Gpcond
\brief  Initialiser une condition
\param  cond adresse du condition
\return 0 OK  1 condition inexistante

*/


extern int PhilCondInit(PhilCond* cond);


/*!\ingroup Gpcond
\brief  Destruction de la condition
\param  cond adresse de la condition
\return un code d'erreur :
	- 0 OK  
	- 1 la condition  est inexistante ou non initialise
	- 2 des threads attendaient sur la condition au moment de la destruction

S'il y a des threads en attente, ils sont rveills et la destruction leur est signale.

*/


extern int PhilCondDestroy(PhilCond* cond);


/*!\ingroup Gpcond
\brief  Signaler une condition
\param  cond adresse de condition
\return un code d'erreur :
	- 0 OK  
	- 1 condition inexistante ou non initialise


	Le thread en attente  le + prioritaire est rveill.

*/


extern int PhilCondSignal(PhilCond* cond);

/*!\ingroup Gpcond
\brief  Signaler une condition
\param  cond adresse de condition
\return un code d'erreur :
	- 0 OK  
	- 1 condition inexistante ou non initialise


	Tous les  threads en attente sont rveills.

*/


extern int PhilCondBroadcast(PhilCond* cond);

/*!\ingroup Gpcond
\brief  Attendre  une condition
\param  cond adresse de condition
\param  mutex verrou  librer au blocage et  reprendre au rveil
\return un code d'erreur :
	- 0 OK 
	- 1 la condition est inexistante ou non initialise
	- 2 la condition  est dtruite ( PhilCondDestroy() ) durant l'attente 

	On peut sortir de l'attente  la suite d'un PhilCondSignal(), un PhilCondBroadcast() ou une destruction de la condition PhilCondDestroy().
*/


extern int PhilCondWait(PhilCond* cond, PhilMutex* mutex);
/*!\ingroup Gpcond
\brief  Attendre  une condition pour une dure borne
\param  cond adresse du condition
\param  mutex verrou  librer au blocage et  reprendre au rveil
\param us dure maximum d'attente en microsecondes
\return un code d'erreur :
	- 0 OK 
	- 1 la condition est inexistante ou non initialise
	- 2 la condition  est dtruite ( PhilCondDestroy() ) durant l'attente 
	- 3 dure limite atteinte 

	On peut sortir de l'attente  la suite d'un PhilCondSignal(), un PhilCondBroadcast(), d'une destruction de la condition PhilCondDestroy() ou de l'puisement du dlai.
*/

extern int PhilCondTimedWait(PhilCond* cond, PhilMutex*  mutex,   long int us);


/*!\ingroup Gpcond
\brief  Gestion de l'heure 
\return la date en microseconde
	
	L'origine des temps sera le dmarrage du programme PHIL

*/

      
extern long int PhilNow();


/*!\ingroup Gpinit
\brief  Initialisation de PHIL
\return un code d'erreur :
	- 0 OK  
	- >0 erreur  l'initialisation

*/
extern int PhilInitialize( );


/*!\ingroup Gpinit
\brief  Terminaison de PHIL
\return un code d'erreur :
	- 0 OK  
	- >0 erreur  la terminaison

	Le thread principal attend la terminaison de tous les threads joignables. Les threads dtachs sont ignors.

*/
extern int PhilTerminate();

/*!\ingroup Gpinit
\brief  Terminaison  fore de PHIL
\param mess le message d'erreur

	Un  thread aborte le programme PHIL.

*/
extern void  PhilAbort(char* mess);
#endif
